@@ -41,6 +41,7 @@ extern PyMethodDef pygtk_functions[];
 extern PyMethodDef pygdk_functions[];
 
 PyObject *PyGtkDeprecationWarning;
+PyObject *PyGtkWarning;
 
 static struct _PyGtk_FunctionStruct functions = {
     VERSION,
@@ -49,6 +50,24 @@ static struct _PyGtk_FunctionStruct functions = {
     pygdk_rectangle_from_pyobject,
 };
 
+static void
+sink_gtkwindow(GObject *object)
+{
+    if (object->ref_count == 1
+	&& GTK_WINDOW(object)->has_user_ref_count) {
+	g_object_ref(object);
+    }
+}
+
+static void
+sink_gtkinvisible(GObject *object)
+{
+    if (object->ref_count == 1
+	&& GTK_INVISIBLE(object)->has_user_ref_count) {
+	g_object_ref(object);
+    }
+}
+
 static void
 sink_gtkobject(GObject *object)
 {
@@ -58,21 +77,35 @@ sink_gtkobject(GObject *object)
     }
 }
 
+static void
+_pygtk_log_func(const gchar *log_domain,
+                GLogLevelFlags log_level,
+                const gchar *message,
+                gpointer user_data)
+{
+    PyGILState_STATE state;
+
+    state = pyg_gil_state_ensure();
+    PyErr_Warn(PyGtkWarning, (char *) message);
+    pyg_gil_state_release(state);
+}
+
 static gboolean
 python_do_pending_calls(gpointer data)
 {
     gboolean quit = FALSE;
+    PyGILState_STATE state;
+
+    state = pyg_gil_state_ensure();
 
-    pyg_block_threads();
     if (PyErr_CheckSignals() == -1) {
 	PyErr_SetNone(PyExc_KeyboardInterrupt);
 	quit = TRUE;
     }
-    pyg_unblock_threads();
-
     if (quit && gtk_main_level() > 0)
 	gtk_main_quit();
     
+    pyg_gil_state_release(state);
     return TRUE;
 }
 
@@ -90,6 +123,8 @@ init_gtk(void)
     init_pygobject();
     g_assert(pygobject_register_class != NULL);
 
+    pygobject_register_sinkfunc(GTK_TYPE_WINDOW, sink_gtkwindow);
+    pygobject_register_sinkfunc(GTK_TYPE_INVISIBLE, sink_gtkinvisible);
     pygobject_register_sinkfunc(GTK_TYPE_OBJECT, sink_gtkobject);
 
     /* set the default python encoding to utf-8 */
@@ -98,7 +133,20 @@ init_gtk(void)
     /* initialise GTK ... */
     av = PySys_GetObject("argv");
     if (av != NULL) {
-	argc = PyList_Size(av);
+	if (!PyList_Check(av)) {
+	    PyErr_Warn(PyExc_Warning, "ignoring sys.argv: it must be a list of strings");
+	    av = NULL;
+	} else {
+	    argc = PyList_Size(av);
+	    for (i = 0; i < argc; i++)
+		if (!PyString_Check(PyList_GetItem(av, i))) {
+		    PyErr_Warn(PyExc_Warning, "ignoring sys.argv: it must be a list of strings");
+		    av = NULL;
+		    break;
+		}
+	}
+    }
+    if (av != NULL) {
 	argv = g_new(char *, argc);
 	for (i = 0; i < argc; i++)
 	    argv[i] = g_strdup(PyString_AsString(PyList_GetItem(av, i)));
@@ -114,14 +162,18 @@ init_gtk(void)
 	    g_free(argv);
 	}
 	PyErr_SetString(PyExc_RuntimeError, "could not open display");
-	/* set the LC_NUMERIC locale back to "C", as Python requires
+	/* set the LC_NUMERIC locale back to "C", as Python < 2.4 requires
 	 * that it be set that way. */
+#if PY_VERSION_HEX < 0x020400F0
 	setlocale(LC_NUMERIC, "C");
+#endif
 	return;
     }
-    /* set the LC_NUMERIC locale back to "C", as Python requires that
+    /* set the LC_NUMERIC locale back to "C", as Python < 2.4 requires that
      * it be set that way. */
+#if PY_VERSION_HEX < 0x020400F0
     setlocale(LC_NUMERIC, "C");
+#endif
     if (argv != NULL) {
 	PySys_SetArgv(argc, argv);
 	for (i = 0; i < argc; i++)
@@ -198,7 +250,7 @@ init_gtk(void)
     pygdk_register_classes(d);
     pygdk_add_constants(m, "GDK_");
       /* Add predefined atoms */
-#define add_atom(name) PyModule_AddObject(m, #name, PyGdkAtom_New(GDK_##name))
+#define add_atom(name) PyModule_AddObject(m, #name, PyString_FromString(gdk_atom_name((GDK_##name))))
 
     add_atom(SELECTION_PRIMARY);
     add_atom(SELECTION_SECONDARY);
@@ -217,6 +269,11 @@ init_gtk(void)
     add_atom(SELECTION_TYPE_WINDOW);
     add_atom(SELECTION_TYPE_STRING);
 #undef add_atom
+
+    PyGtkWarning = PyErr_NewException("gtk.GtkWarning", PyExc_Warning, NULL);
+    PyDict_SetItemString(d, "Warning", PyGtkWarning);
+    g_log_set_handler("Gtk", G_LOG_LEVEL_CRITICAL|G_LOG_LEVEL_WARNING,
+                      _pygtk_log_func, NULL);
     
     gtk_timeout_add(100, python_do_pending_calls, NULL);
 }
